classdef CChannel < handle
    %CCHANNEL NiHu.Lab class for Channel handling
    
    properties (SetAccess = immutable)
        PhysChNr        % physical channel number
        Inp             % input object handle
    end
    
    properties (SetObservable = true)
        isSelected      % logical
        isReference     % logical
        Name            % Channel name string
        Dir             % Channel direction
        Sensor          % Sensor object handle
        AmpCh           % Amplifier Channel object handle
        Color
    end
    
    methods
        function obj = CChannel(nr, sel, ref, name, dir, inp, sens, amp, color)
            % CCHANNEL Constructor
            obj.PhysChNr = nr;
            obj.isSelected = sel;
            obj.isReference = ref;
            obj.Name = name;
            obj.Dir = dir;
            obj.Inp = inp;
            obj.Sensor = sens;
            obj.AmpCh = amp;
            obj.Color = color;
        end
        
        function opmode = OpMode(obj)
            %OPMODE return operation mode of the channel(s) (AC or ICP)
            opmode = repmat(COperationMode.AC, size(obj));
            for i = 1 : numel(obj)
                if obj(i).isSelected && obj(i).Sensor.requireICP && ~obj(i).AmpCh.Amplifier.provideICP
                    opmode(i) = COperationMode.ICP;
                end
            end
        end
        
        function amp = voltage2eu(obj)
            %VOLTAGE2EU convert voltage to engineering unit
			% AMP = VOLTAGE2EU(OBJ) returns a constant converting measured voltage to actual EU.
			% The constant takes the actual sensitivity and the preamplifier
			% amplification into account.
            amp = ones(size(obj));
            for i = 1 : numel(obj)
                amp(i) = 1 / (obj(i).AmpCh.Amplification * obj(i).Sensor.ActualSensitivity);
            end
        end
        
        function amp = PreAmplification(obj)
            %PREAMPLIFICATION return preamplification of the channel(s)
            amp = ones(size(obj));
            for i = 1 : numel(obj)
                amp(i) = obj(i).AmpCh.Amplification;
            end
        end
        
        function qt = getQuant(obj)
            %GETQUANT return measured quantity in object format
            qt = repmat(CQuantity.Voltage, size(obj));
            for i = 1 : numel(obj)
                qt(i) = obj(i).Sensor.Quantity;
            end
        end
        
        function qt = getQuantText(obj)
            %GETQUANTTEXT returner measured quantity in textual format
            qt = cell(size(obj));
            for i = 1 : numel(obj)
                qt{i} = upper(obj(i).Sensor.Quantity.char());
            end
        end
        
        function eu = getEUText(obj)
            %GETEUTEXT return measured engineergin unit in textual format
            eu = cell(size(obj));
            for i = 1 : numel(obj)
                eu{i} = obj(i).Sensor.Quantity.EU();
            end
        end
        
        function r = getRange(obj)
            %GETRANGE return channel range
            r = zeros(size(obj));
            for i = 1 : numel(obj)
                r(i) = obj(i).Inp.Mod.Range;
            end
        end
        
        function r = getModule(obj)
            %GETMODULE return channel module
            r = cell(size(obj));
            for i = 1 : numel(obj)
                r(i) = obj(i).Inp.Mod.getModel();
            end
        end
        
        function [t, h, f] = toLatexTabular(obj)
            %TOLATEXTABULAR convert channel to LaTeX tabular
            if (numel(obj) > 1)
                t = cell(numel(obj),1);
                for i = 1 : numel(obj)
                    [t{i}, h, f] = obj(i).toLatexTabular();
                end
            else
                f = 'llll';
                h = 'Name & Dir. & Sensor & Amplifier';
                t = sprintf('%s & %s & %s & %s', ...
                    obj.Name, ...
                    obj.Dir.char, ...
                    obj.Sensor.Name, ...
                    obj.AmpCh.Name);
            end
        end
    end % of methods
end % of classdef
