classdef CDAQmxUSB < CNi9234Hardware
    %CDAQmx Simulated hardware class
    %   This class implements a NI CDAQmx driver based hardware
    
    properties (Constant = true)
        DMADepth = 2e6;         % Depth of input and output DMA on HOST side
        TriggerBlockDuration = .1; % Software trigger block size in sec
    end
    
    properties
        DeviceProperties
        TriggerBlockSize
        
        TaskHandleIn
        
        InputModules     % vector of CModule objects
        InputChannelNames
        OutputModules     % vector of CModule objects
        OutputChannelNames

        nChIn         % number of input channels
        
        nPretrigSamples     % number of PreTrigger Samples
        nPosttrigSamples    % number of PostTrigger Samples
        nSamples            % total number of samples to acquire
        
        ImpactBuffer        % internal Buffer for impact testing hw trigger
        ImpactBufferIndex   % index pointer of circular buffer
    end
    
    properties (SetObservable = true)
        nSamplesInBuffer    % number of samples in buffer
    end
    
    methods
        % Constructor
        function obj = CDAQmxUSB()
            obj = obj@CNi9234Hardware();
            
            devnames = NiHuDAQmx('get_device_names');
            devnames = regexp(devnames, 'Dev\d', 'match');
            if isempty(devnames)
                error('nem talaltam');
            end
            obj.DeviceProperties = NiHuDAQmx('get_device_properties', devnames{1});
            
            im = CModule.empty;
            in = [];
            
            mn = obj.DeviceProperties.DeviceName;
            mod = CModule.FromSN(obj.DeviceProperties.SerialNumber);
            if (mod.isInput)
                    im = [ im; mod ];
                    in = [ in; strcat(mn, '/ai', cellstr(num2str((1:im(end).nChannels)'-1))) ];
            else
            end
            
            obj.InputModules = im;
            obj.InputChannelNames = in;
            obj.OutputModules = CModule.empty;
            obj.OutputChannelNames = [];
        end
        
        function delete(obj)
        end
        
        function status = MonitorInit(obj, Channels)
            % Initialise Monitor Module
            
            obj.TaskHandleIn = NiHuDAQmx('create_task', 'Monitor10Task');
            
            ChIdx = find(cell2mat({Channels.isSelected}));
            obj.nChIn = length(ChIdx);
            
            try
                for c = 1 : obj.nChIn
                    idx = ChIdx(c);
                    icp = Channels(idx).Sensor.requireICP &&...
                        Channels(idx).Inp.Mod.provideICP;
                    NiHuDAQmx('create_input_channel',...
                        obj.TaskHandleIn,...
                        obj.InputChannelNames{idx},...
                        Channels(idx).Inp.Mod.Range * [-1 1], ...
                        icp);
                end
            catch ME
                disp(ME.message);
                NiHuDAQmx('clear_task',  obj.TaskHandleIn);
            end
            
            status = 0;
        end
        
        function [status, starttime] = MonitorStart(obj, Fs, nSamp)
            % Start Monitoring
            
            NiHuDAQmx('configure_timing', obj.TaskHandleIn, Fs, true, uint64(5*nSamp));
            NiHuDAQmx('start_task', obj.TaskHandleIn);
            
            starttime = now();
            status = 0;
        end
        
        function [status, voltage, elemRemaining] = MonitorAcq(obj, nSamples)
            % Acquire samples during monitoring
            
            [voltage, nRead] = NiHuDAQmx('read', obj.TaskHandleIn, int32(nSamples));
            if (nRead ~= nSamples)
                error('Could not read enough samples');
            end
            
            elemRemaining = 0;
            status = 0;
        end
        
        function [status, rest] = MonitorStop(obj)
            % Stop Monitoring
            
            NiHuDAQmx('stop_task', obj.TaskHandleIn);
            
            rest = zeros(0, obj.nChIn);
            status = 0;
        end
        
        function status = MonitorClose(obj)
            % Close Monitor Module
            
            NiHuDAQmx('clear_task', obj.TaskHandleIn);
            
            status = 0;
        end
        
        function status = TransferInit(obj, InputChannels, OutputChannels)
            %TRANSFERINIT Initialise DAQmx hardware for Transfer module
            
            obj.TaskHandleIn = NiHuDAQmx('create_task', 'MonitorInputTask');
            obj.TaskHandleOut = NiHuDAQmx('create_task', 'MonitorOuptutTask');
            
            ChInIdx = find(cell2mat({InputChannels.isSelected}));
            obj.nChIn = length(ChInIdx);
            ChOutIdx = find(cell2mat({OutputChannels.isSelected}));
            obj.nChOut = length(ChOutIdx);
            
            try
                for c = 1 : obj.nChIn
                    idx = ChInIdx(c);
                    icp = InputChannels(idx).Sensor.requireICP &&...
                        InputChannels(idx).Inp.Mod.provideICP;
                    NiHuDAQmx('create_input_channel',...
                        obj.TaskHandleIn,...
                        obj.InputChannelNames{idx},...
                        InputChannels(idx).Inp.Mod.Range * [-1 1], ...
                        icp);
                end
                
                for c = 1 : obj.nChOut
                    idx = ChOutIdx(c);
                    NiHuDAQmx('create_output_channel',...
                        obj.TaskHandleOut,...
                        obj.OutputChannelNames{idx},...
                        OutputChannels(idx).Inp.Mod.Range * [-1 1]);
                end
            catch ME
                disp(ME.message);
                NiHuDAQmx('clear_task',  [obj.TaskHandleIn obj.TaskHandleOut]);
            end
            
            status = 0;
        end
        
        function [status, y]= TransferAcq(obj, nPre, nAverage, nPost, Fs, x)
            %TRANSFERACQ Acquire transfer functions
            
            nTot = nPre + nAverage + nPost;
            nSamp = size(x,1);
            
            NiHuDAQmx('configure_timing', [obj.TaskHandleIn obj.TaskHandleOut], Fs, false,...
                uint64(nSamp*nTot));
            
            if size(x,2) == 1
                x = repmat(x, 1, nTot);
            end
            num_written = NiHuDAQmx('write', obj.TaskHandleOut,...
                repmat(x(:), 1, obj.nChOut));
            
            NiHuDAQmx('start_task', [obj.TaskHandleIn obj.TaskHandleOut]);
            
            [y, num_read] = NiHuDAQmx('read', obj.TaskHandleIn,...
                int32(nSamp*nTot));
            
            y = reshape(y, nSamp, nTot, []);
            y = y(:,nPre+(1:nAverage), :);
            
            obj.nSamplesInBuffer = 0;
            
            NiHuDAQmx('stop_task', [obj.TaskHandleIn obj.TaskHandleOut]);
            
            status = 0;
        end
        
        function status = TransferClose(obj)
            % Close CDAQmx in Transfer Module
            NiHuDAQmx('clear_task', [obj.TaskHandleIn obj.TaskHandleOut]);
            status = 0;
        end
        
        
        function status = ImpactInit(obj, InputChannels)
            %IMPACTINIT Initialise DAQmx hardware for Impact module
            
            obj.TaskHandleIn = NiHuDAQmx('create_task', 'ImpactTask');
            
            ChInIdx = find(cell2mat({InputChannels.isSelected}));
            obj.nChIn = length(ChInIdx);
            
            try
                for c = 1 : obj.nChIn
                    idx = ChInIdx(c);
                    icp = InputChannels(idx).Sensor.requireICP &&...
                        InputChannels(idx).Inp.Mod.provideICP;
                    NiHuDAQmx('create_input_channel',...
                        obj.TaskHandleIn,...
                        obj.InputChannelNames{idx},...
                        InputChannels(idx).Inp.Mod.Range * [-1 1], ...
                        icp);
                end
            catch ME
                disp(ME.message);
                NiHuDAQmx('clear_task',  obj.TaskHandleIn);
            end
            
            status = 0;
        end
        
        function status = ImpactArm(obj, nSamples, nPretrigSamples, Fs)
            % Arm impact testing (collect pretrigger samples)
            obj.TriggerBlockSize = round(obj.TriggerBlockDuration * Fs);
            
            obj.nSamples = nSamples;
            obj.nPretrigSamples = nPretrigSamples;
            obj.ImpactBuffer = zeros(obj.nSamples, obj.nChIn);
            
            %configure timing
            NiHuDAQmx('configure_timing', obj.TaskHandleIn, Fs, true,...
                uint64(nSamples));
            
            %start acquisition
            NiHuDAQmx('start_task', obj.TaskHandleIn);
            
            % read pretrig samples
            [y, num_read] = NiHuDAQmx('read', obj.TaskHandleIn, int32(nPretrigSamples));
                        
            index = 1:obj.nPretrigSamples;
            obj.ImpactBuffer(index, :) = y;
            obj.ImpactBufferIndex = obj.nPretrigSamples;
            status = 0;
        end
        
        function [status, isTriggered] = ImpactWaitForTrigger(obj, trigLevel, trigCh)
            % read a block from the hardware
            index = obj.ImpactBufferIndex+(1:obj.TriggerBlockSize);
            [block, num_read] = NiHuDAQmx('read', obj.TaskHandleIn,...
                int32(obj.TriggerBlockSize));
            
            % determine whether trigger event has occured
            indTrig = find(abs(block(:, trigCh)) > trigLevel, 1, 'first');
            isTriggered = ~isempty(indTrig);
            if isTriggered
                % number of samples to read
                obj.nPosttrigSamples =...
                    obj.nSamples - obj.nPretrigSamples - (size(block,1) - indTrig + 1);
                if obj.nPosttrigSamples < 0
                    block = block(1 : end-abs(obj.nPosttrigSamples), :);
                    index = index(1 : end-abs(obj.nPosttrigSamples));
                    obj.nPosttrigSamples = 0;
                end
            end
            obj.ImpactBuffer(mod(index-1, obj.nSamples)+1, :) = block;
            obj.ImpactBufferIndex = obj.ImpactBufferIndex + size(block,1);
            status = 0;
        end
        
        function [status, y]= ImpactAcq(obj)
            % Read post trigger samples and stop acquisition
            
            if obj.nPosttrigSamples > 0
                
                % read post trigger samples into block
                index = obj.ImpactBufferIndex + (1:obj.nPosttrigSamples);
                [block, num_read] = NiHuDAQmx('read', obj.TaskHandleIn, int32(obj.nPosttrigSamples));
                
                % copy block into Buffer and adjust ImpactBufferIndex
                obj.ImpactBuffer(mod(index-1, obj.nSamples)+1,:) = block;
                
                obj.ImpactBufferIndex = obj.ImpactBufferIndex + obj.nPosttrigSamples;
            end
            
            obj.ImpactBufferIndex = mod(obj.ImpactBufferIndex-1, obj.nSamples)+1;
            
            % rotate content of buffer
            y = circshift(obj.ImpactBuffer, -obj.ImpactBufferIndex);
            
            % Stop acquisition and flush DMA
            NiHuDAQmx('stop_task', obj.TaskHandleIn);
            obj.nSamplesInBuffer = 0;
            
            status = 0;
        end
        
        function status = ImpactClose(obj)
            % Close impact testing module
            NiHuDAQmx('clear_task', obj.TaskHandleIn);
            status = 0;
        end
    end % of methods
end % of class CDAQmx
