classdef CGraphDataView < CGuiView & CPanel
    %CGRAPHDATAVIEW View class of CGraphData
    %
    % See also: CGraphData
    
    properties (Access = private)
        hAxis       % Figure Axis handle
        hLines      % Figure Line Handles
        hContextMenu
        ChIndices   % actual Data Indices
        ChIDs       % actual Channel IDs
        isBlocked   % boolean indicating whether drawing is blocked
        
        listLanguageChange
    end
    
    methods (Access = public)
        function obj = CGraphDataView(parent, hData, title)
            %CGRAPHDATAVIEW constructor
            
            obj = obj@CPanel(parent, title);
            obj = obj@CGuiView(hData, {'yData'});
            obj.isBlocked = true;
            obj.ChIndices = 1;
            obj.ChIDs = obj.hData.ChIDs(obj.ChIndices);
            obj.InitObject();
            obj.listLanguageChange = addlistener(CBabel.getInstance(), 'currentLanguage', 'PostSet', @(src,event)languageChangeCallback(obj));
        end
        
        function delete(obj)
            %DELETE destructor
            
            delete(obj.listLanguageChange);
        end
        
        function languageChangeCallback(obj)
            %LANGUAGECHANGECALLBACK called when language is changed
            
            obj.SetLabels();
        end
        
        function setScale(obj, axis, value)
            %SETSCALE change scale of all axes to value
            
            for i = 1 : length(axis)
                set(obj.hAxis, [axis(i) 'scale'], value);
            end
        end
        
        function block(obj)
            %BLOCK block the user interface

            obj.isBlocked = true;
        end
        
        function unblock(obj)
            %UNBLOCK unblock the user interface
            
            obj.isBlocked = false;
        end
        
        function refresh(obj)
            %REFRESH Called when the GUI needs to be updated
            
            if ~obj.isBlocked
                obj.PlotLines();
            end
        end
        
        function UpdateChannels(obj)
            %UPDATECHANNELS called when the channels are updated

            [~, newindices] = ismember(obj.ChIDs, obj.hData.ChIDs);
            newindices = newindices(newindices ~= 0);
            if isempty(newindices)
                obj.ChIndices = 1;
            else
                [~, obj.ChIndices] = obj.CheckConsistency(newindices);
            end
            obj.ChIDs = obj.hData.ChIDs(obj.ChIndices);
            
            obj.PlotLines();
            obj.PlotLegend();
            obj.SetLabels();
        end
        
        function LegendButtonPressCallback(obj)
            %LEGENDBUTTONPRESSCALLBACK Callback function of legend selection event
            
            % no selection if only one channel exists
            nCh = length(obj.hData.ChNames);
            if nCh == 1
                return;
            end
            
            % open selection figure
            f = figure('Name', 'Choose Channel',...
                'NumberTitle', 'off',...
                'Units', 'Pixels',...
                'MenuBar', 'None',...
                'Resize', 'off',...
                'WindowStyle', 'modal',...
                'Visible', 'off',...
                'Color', get(0,'defaultUicontrolBackgroundColor'),...
                'KeyPressFcn', @(src,event)GuiLayout.CloseToEscapeCallback(event));
            % create selection box with appropriate default selection
            l = uicontrol('parent', f,...
                'style', 'listbox', ...
                'Min', 1,...
                'Max', 3,...
                'units', 'pixels', ...
                'KeyPressFcn', @(src,event)GuiLayout.CloseToEscapeCallback(event), ...
                'String', obj.hData.ChNames,...
                'Value', obj.ChIndices);
            p = get(l, 'extent') + CSetupClass.ListBoxExtentCorrection;
            p(3) = max(p(3), CSetupClass.legendWidth);
            set(l, 'position', p);
            % create OK button
            b = uicontrol('parent', f,...
                'style', 'pushButton', ...
                'position', [0 0 p(3), CSetupClass.itemHeight], ...
                'KeyPressFcn', @(src,event)GuiLayout.CloseToEscapeCallback(event), ...
                'Tag', 'DONE', ...
                'String', CBabel.getInstance().getText('DONE'),...
                'Callback', @(src,event)ChSelectDoneCallback(obj,f,l));
            GuiLayout.ApplyLayout(f, {{l, [0 0]};{b, [0 0]}});
            GuiLayout.MoveGuiRel(f, 'cursor');
            movegui(f, 'onscreen');
            set(f, 'Visible', 'on');
        end
        
        function ChSelectDoneCallback(obj, hFig, hSelect)
            %CHSELECTDONECALLBACK Callback function called when legend selection is done
            
            % the selected indices
            index = get(hSelect, 'Value');
            
            % check if indices refer to consistent channels
            if ~CheckConsistency(obj, index)
                errordlg(CBabel.getInstance().getText('LEGEND_EU_CONFLICT_ERROR'), ...
                    CBabel.getInstance().getText('DISPLAY_ERROR'), ...
                    'modal');
                return;
            end
            
            % close the selection window
            close(hFig);
            
            obj.ChIndices = index;
            obj.ChIDs = obj.hData.ChIDs(obj.ChIndices);
            obj.PlotLines();
            obj.SetLabels();
            obj.PlotLegend();
        end
        
        function AxisTableCallback(obj, src, event)
            %AXISTABLECALLBACK called when a value is changed in the table
            
            row = event.Indices(1);
            col = event.Indices(2);
            D = get(src, 'Data');
            axis = get(src, 'RowName');
            axis = axis{row};
            switch col
                case {1,2}
                    lim = cell2mat(D(row,1:2));
                    SetLim(obj.hAxis, [axis 'Lim'], lim);
                    D{row,4} = 'manual';
                case 3
                    l = D{row,3};
                    set(obj.hAxis, [axis 'Scale'], l);
                case 4
                    set(obj.hAxis, [axis 'LimMode'], D{row,4});
                    D(row,1:2) = num2cell(get(obj.hAxis, [axis 'Lim']));
            end
            set(src, 'Data', D);
        end
        
        function AxisButtonPressCallback(obj)
            %AXISBUTTONPRESSCALLBACK called when the user clicks inside the axis
            
            switch(get(gcbf, 'SelectionType'))
                case 'alt' % right click
                    XL = get(obj.hAxis, 'XLim');
                    YL = get(obj.hAxis, 'YLim');
                    ScaleStr = {'linear', 'log'};
                    ModeStr = {'auto', 'manual'};
                    [~, iX] = ismember(get(obj.hAxis, 'XScale'), ScaleStr);
                    [~, iY] = ismember(get(obj.hAxis, 'YScale'), ScaleStr);
                    [~, jX] = ismember(get(obj.hAxis, 'XLimMode'), ModeStr);
                    [~, jY] = ismember(get(obj.hAxis, 'YLimMode'), ModeStr);
                    
                    MainFigure = figure('Name', 'Set Axes Properties',...
                        'NumberTitle', 'off',...
                        'Units', 'Pixels',...
                        'MenuBar', 'None',...
                        'Resize', 'off',...
                        'WindowStyle', 'modal',...
                        'Visible', 'off',...
                        'Color', get(0,'defaultUicontrolBackgroundColor'), ...
                        'KeyPressFcn', @(src,event)GuiLayout.CloseToEscapeCallback(event), ...
                        'position', [get(0, 'PointerLocation')-[274 56]/2, [274 56]]);
                    
                    t = uitable('parent', MainFigure,...
                        'ColumnName', {'min', 'max', 'type', 'auto'}, ...
                        'ColumnFormat', {'numeric', 'numeric', ScaleStr, ModeStr}, ...
                        'ColumnEditable', true(1,4), ...
                        'ColumnWidth', {70 70 50 50}, ...
                        'RowName', {'y', 'x'}, ...
                        'Data', {YL(1), YL(2), ScaleStr{iY}, ModeStr{jY};
                        XL(1), XL(2), ScaleStr{iX}, ModeStr{jX}}, ...
                        'CellEditCallback', @(src,event)AxisTableCallback(obj, src, event));
                    set(MainFigure, 'visible', 'on');
                    e = get(t, 'extent');
                    set(t, 'position', [0 0 e(3:4)]);
                    set(MainFigure, 'position', [get(0, 'PointerLocation')-e(3:4)/2, e(3:4)]);
                case 'open' % double click (copy figure)
                    fNew = figure;
                    h = get(get(obj.hAxis, 'Parent'), 'Children');
                    % get lengend as it cannot be copied
                    hl = findobj(h, 'Tag', 'legend');
                    h = setdiff(h, hl);
                    
                    copyobj(h, fNew);
                    legend(h);
                    set(get(fNew, 'Children'), 'ButtonDownFcn', '');
                otherwise
            end
        end
        
        function PlayBackCallback(obj)
            %PLAYBACKCALLBACK Called when the user plays a curve
            x = get(gco, 'xData');
            y = get(gco, 'yData');
            XL = get(obj.hAxis, 'xLim');
            ind = x > XL(1) & x < XL(2);
            soundsc(y(ind), 1/(x(2)-x(1)));
        end
    end % of public methods
    
    methods (Access = private)
        function InitObject(obj)
            obj.hContextMenu = uicontextmenu('parent', ancestor(obj.hPanel, 'Figure'));
            uimenu(obj.hContextMenu, 'Label', 'PLayBack', 'Callback', @(src,event)PlayBackCallback(obj))
            obj.hAxis = axes('parent', obj.hPanel,...
                'ActivePositionProperty', 'outerposition', ...
                'NextPlot', 'ReplaceChildren', ...
                'ButtonDownFcn', @(src,event)AxisButtonPressCallback(obj));
            hold(obj.hAxis, 'on');
            GuiLayout.ApplyLayout(obj.hPanel, {{obj.hAxis, [-1 -1]}});
        end
        
        function SetLabels(obj)
            ind = obj.ChIndices(1);
            xlabel(obj.hAxis, [CBabel.getInstance().getText(obj.hData.xQuant), ' [' obj.hData.xEU ']']);
            ylabel(obj.hAxis, [CBabel.getInstance().getText(obj.hData.yQuant{ind}), ' [' obj.hData.yEU{ind} ']']);
        end
        
        function PlotLines(obj)
            for i = 1 : length(obj.ChIndices)
                idx = obj.ChIndices(i);
                if i > length(obj.hLines)
                    obj.hLines(i,1) = plot(obj.hAxis,...
                        obj.hData.xData, obj.hData.yData(:,idx), ...
                        'Color', obj.hData.Colors(idx,:), ...
                        'UIContextMenu', obj.hContextMenu, ...
                        'DisplayName', obj.hData.ChNames{idx});
                else
                    set(obj.hLines(i), 'xData', obj.hData.xData,...
                        'yData', obj.hData.yData(:,idx), ...
                        'Color', obj.hData.Colors(idx,:), ...
                        'DisplayName', obj.hData.ChNames{idx});
                end
            end
            delete(obj.hLines(i+1:end,:));
            obj.hLines = obj.hLines(1:i,:);
        end
        
        function PlotLegend(obj)
            l = legend(obj.hAxis, obj.hLines);
            set(l, 'ButtonDownFcn', @(src,event)LegendButtonPressCallback(obj), ...
                'FontSize', 8);
        end
        
        function [isConsistent, index] = CheckConsistency(obj, index)
            [a, ~, b] = unique(obj.hData.yEU(index));
            isConsistent = length(a) == 1;
            index = index(b == 1);
        end
    end % of private methods
end % of class CGraphDataView

function SetLim(hAxes, prop, values)
if ~any(isnan(values))
    values = sort(values);
    if values(2) > values(1)
        set(hAxes, prop, values);
    end
end
end
