classdef CPlayrec < CHardware
    %CPLAYREC Hardware class to provide interface to SoundCards using Playrec
    %   This class implements an interface between NiHu.Lab and SoundCards using Playrec
    
    properties (Constant = true)
        DMADepth = 2e6;
        TriggerBlockDuration = .1; % Software trigger block size in sec
    end
    
    properties
        InputModules        % vector of CModule input objects
        OutputModules       % vector of CModule output objects
        
        playDevice          % the ID of the Playrec device to be used for sample output
        recDevice           % the ID of the Playrec device to be used for sample input
        playChanList        % a 1xN vector containing the channels on which the playBuffer samples should be output
        recChanList         % a row vector containing the channel numbers of all channels to be recorded
        t                   % Timer object
        lastReadPage        % Last read Playrec memory page
        nSamples            % Number of samples
        
        nPretrigSamples     % number of PreTrigger Samples
        nPosttrigSamples    % number of PostTrigger Samples
        
        TriggerBlockSize
        
        ImpactBuffer        % internal Buffer for impact testing hw trigger
        ImpactBufferIndex   % index pointer of circular buffer

        defaultFs
    end
    
    properties (Access = private, Dependent = true)
        nInCh           % number of input channels
        nOutCh          % number of output channels
    end
    
    properties (SetObservable = true)
        nSamplesInBuffer
    end
    
    methods
        % Constructor
        function obj = CPlayrec()
            obj.InputModules = CModule.empty(0,1);
            obj.OutputModules = CModule.empty(0,1);
            d = playrec('getDevices');
            if isempty(d)
                error('No PlayRec Devices found');
            end
            for i = 1 : length(d)
                switch d(i).name
                    case 'ASIO E-MU Tracker Pre | USB'
                        obj.InputModules(end+1) = CModule.TrackerPreI;
                        obj.OutputModules(end+1) = CModule.TrackerPreO;
                    case 'UA-25'
                        obj.InputModules(end+1) = CModule.UA25I;
                        obj.OutputModules(end+1) = CModule.UA25O;
                    case 'Yamaha USB ASIO'
                        obj.InputModules(end+1) = CModule.Audiogram3I;
                        obj.OutputModules(end+1) = CModule.Audiogram3O;
                    case 'ASIO Fireface'
                        obj.InputModules(end+1) = CModule.FirefaceI;
                        obj.OutputModules(end+1) = CModule.FirefaceO;
                    otherwise
                        error('Unknown PlayRec hardware %s', d(i).name);
                end
            end
            
            obj.playDevice = d(1).deviceID;     % TODO: depends on selection, has to move to Init functions
            obj.recDevice = d(1).deviceID;      % TODO: depends on selection, has to move to Init functions
            obj.defaultFs = d(1).defaultSampleRate;
            
            obj.playChanList = [];
            obj.recChanList = [];
            obj.lastReadPage = -1;
        end
        
        function fsRange = getFsRange(obj)  % TODO
            fsRange = obj.defaultFs;
        end
        
        function status = MonitorInit(obj, Channels)
            % Initialise Monitor Module
            
            ChSel = cell2mat({Channels.isSelected});
            obj.recChanList = find(ChSel);
            
            status = 0;
        end
        
        function [status, starttime] = MonitorStart(obj, Fs, nSamples)
            % Start Monitoring
            
            dT = nSamples/Fs;
            obj.nSamples = nSamples;        % needed for MonitorStop
            playrec('init', Fs, obj.playDevice, obj.recDevice);    % Input Arguments: (sampleRate, playDevice, recDevice)
            
            obj.t = timer('BusyMode', 'queue',...
                'Period', dT, 'StartDelay', dT/2,...
                'ExecutionMode', 'fixedRate',...
                'StartFcn',@(x,y)TimerCallback(obj, nSamples),...
                'TimerFcn',@(x,y)TimerCallback(obj, nSamples));
            
            start(obj.t);
            starttime = now;
            status = 0;
        end
        
        function [status, voltage] = MonitorAcq(obj, ~)
            % Acquire samples during monitoring
            
            while playrec('getLastFinishedPage') < obj.lastReadPage+1
                disp('waiting');
                pause(.1);
            end
            
            voltage = playrec('getRec', obj.lastReadPage+1);
            disp('acquiring');
            if isempty(voltage) == 0
                obj.lastReadPage = obj.lastReadPage+1;
            end
            
            obj.nSamplesInBuffer = (playrec('getLastFinishedPage')-obj.lastReadPage)*obj.nSamples;
            status = 0;
        end
        
        function [status rest] = MonitorStop(obj)
            % Stop Monitoring
            
            stop(obj.t);
            
            playrec('block');
            remainingPages = playrec('getLastFinishedPage')-obj.lastReadPage;
            
            rest = zeros(obj.nSamples*remainingPages, obj.nInCh);
            for i = (1:remainingPages)
                rest((i-1)*obj.nSamples+1:i*obj.nSamples,:) = playrec('getRec', obj.lastReadPage+i);
            end
            
            obj.lastReadPage = playrec('getLastFinishedPage');
            obj.nSamplesInBuffer = 0;
            
            playrec('delPage');
            playrec('reset');
            delete(obj.t);
            status = 0;
        end
        
        function status = MonitorClose(obj)
            %MONITORCLOSE Close Monitor Module
            
            obj.playChanList = [];
            obj.recChanList = [];
            status = 0;
        end
        
        function status = TransferInit(obj, InputChannels, OutputChannels)
            % Initialise Playrec for Transfer module
            
            obj.recChanList = find(cell2mat({InputChannels.isSelected}));
            obj.playChanList = find(cell2mat({OutputChannels.isSelected}));
            
            status = 0;
        end
        
        function [status y]= TransferAcq(obj, nPre, nAverage, nPost, Fs, x)
            playrec('init', Fs, obj.playDevice, obj.recDevice);
            nTot = nPre + nAverage + nPost;
            nsamples = size(x,1);
            
            if size(x,2) == 1
                x = repmat(x, nTot, 1);
            end
            
            pageNumber = playrec('playrec', repmat(x(:), 1, obj.nOutCh), obj.playChanList, -1, obj.recChanList); % Input Arguments: (playBuffer, playChanList, recDuration, recChanList)
            playrec('block');
            voltage = playrec('getRec', pageNumber);
            voltage = reshape(voltage, nsamples, nTot, obj.nInCh);
            y = voltage(:,nPre+(1:nAverage), :);
            
            obj.lastReadPage = playrec('getLastFinishedPage');
            playrec('delPage');
            playrec('reset');
            status = 0;
        end
        
        function status = TransferClose(obj)
            % Close Playrec in Transfer Module
            
            obj.playChanList = [];
            obj.recChanList = [];
            status = 0;
        end
        
        function status = ImpactInit(obj, Channels)
            % Initialise FPGA for Impact module
            
            ChSel = cell2mat({Channels.isSelected});
            obj.recChanList = find(ChSel);
            
            status = 0;
            
        end % of function ImpactInit
        
        function status = ImpactArm(obj, nSamples, nPretrigSamples, Fs)
            % Arm impact testing (collect pretrigger samples)
            
            obj.nSamples = nSamples;
            obj.nPretrigSamples = nPretrigSamples;
            obj.ImpactBuffer = zeros(obj.nSamples, obj.nInCh);
            obj.TriggerBlockSize = round(obj.TriggerBlockDuration * Fs);
            
            % start acquistion
            playrec('init', Fs, obj.playDevice, obj.recDevice);
            if (obj.nPretrigSamples/Fs)-(obj.TriggerBlockDuration/2) < 0
                StartDelay = obj.nPretrigSamples/Fs/2;
            else
                StartDelay = (obj.nPretrigSamples/Fs)-(obj.TriggerBlockDuration/2);
            end
            obj.t = timer('Period', obj.TriggerBlockDuration,...
                'BusyMode', 'queue',...
                'StartDelay', StartDelay,...
                'ExecutionMode', 'fixedRate',...
                'StartFcn',@(x,y)PlayrecImpTimCbackStFcn(obj),...
                'TimerFcn',@(x,y)PlayrecImpTimCbackTiFcn(obj));
            start(obj.t);
            
            % read pretrig samples
            playrec('block');
            Samples = playrec('getRec', obj.lastReadPage+1);
            obj.lastReadPage = obj.lastReadPage+1;
            obj.nSamplesInBuffer = (size(Samples,1));
            
            % copy samples into buffer and adjust sample index offset
            index = 1:obj.nPretrigSamples;
            obj.ImpactBuffer(index, :) = reshape(Samples, [], obj.nInCh);
            obj.ImpactBufferIndex = obj.nPretrigSamples;
            
            status = 0;
        end % of function ImpactArm
        
        function [status, isTriggered] = ImpactWaitForTrigger(obj, trigLevel, trigCh)
            % Read a block and search for trigger event inside the block
            
            % read a block from the hardware
            index = obj.ImpactBufferIndex+(1:obj.TriggerBlockSize);
            while playrec('getLastFinishedPage') < obj.lastReadPage+1
                disp('waiting');
                pause(obj.TriggerBlockDuration);
            end
            block = playrec('getRec', obj.lastReadPage+1);
            disp('acquiring');
            if isempty(block) == 0
                obj.lastReadPage = obj.lastReadPage+1;
            end
            block = reshape(block, [], obj.nInCh);
            obj.nSamplesInBuffer = (playrec('getLastFinishedPage')-...
                obj.lastReadPage)*obj.nSamples;
            
            % determine whether trigger event has occured
            indTrig = find(abs(block(:, trigCh)) > trigLevel, 1, 'first');
            isTriggered = ~isempty(indTrig);
            if isTriggered
                % number of samples to read
                obj.nPosttrigSamples =...
                    obj.nSamples - obj.nPretrigSamples - (size(block,1) - indTrig + 1);
                if obj.nPosttrigSamples < 0
                    block = block(1 : end-abs(obj.nPosttrigSamples), :);
                    index = index(1 : end-abs(obj.nPosttrigSamples));
                    obj.nPosttrigSamples = 0;
                end
            end
            obj.ImpactBuffer(mod(index-1, obj.nSamples)+1, :) = block;
            obj.ImpactBufferIndex = obj.ImpactBufferIndex + size(block,1);
            status = 0;
        end % of function ImpactWaitForTrigger
        
        function [status y]= ImpactAcq(obj)
            % Read post trigger samples from DMA and stop acquisition
            
            if obj.nPosttrigSamples > 0
                
                % read post trigger samples into block
                index = obj.ImpactBufferIndex + (1:obj.nPosttrigSamples);
                
                % stop acquistion
                while (playrec('getLastFinishedPage')-obj.lastReadPage)*obj.TriggerBlockSize...
                        <= obj.nPosttrigSamples
                    disp('waiting for posttrigsamples');
                    pause(obj.TriggerBlockDuration);
                end
                stop(obj.t);
                
                playrec('block');
                remainingPages = playrec('getLastFinishedPage')-obj.lastReadPage;
                rest = zeros(obj.TriggerBlockSize*remainingPages, obj.nInCh);
                for i = (1:remainingPages)
                    rest((i-1)*obj.TriggerBlockSize+1:i*obj.TriggerBlockSize,:) =...
                        playrec('getRec', obj.lastReadPage+i);
                end
                obj.lastReadPage = obj.lastReadPage+i;
                block = rest(1:obj.nPosttrigSamples, :);
                
                % copy block into Buffer and adjust ImpactBufferIndex
                obj.ImpactBuffer(mod(index-1, obj.nSamples)+1,:) =...
                    reshape(block, [], obj.nInCh);
                
                obj.ImpactBufferIndex = obj.ImpactBufferIndex + obj.nPosttrigSamples;
            end
            
            obj.ImpactBufferIndex = mod(obj.ImpactBufferIndex-1, obj.nSamples)+1;
            
            % rotate content of buffer
            y = circshift(obj.ImpactBuffer, -obj.ImpactBufferIndex);
            
            obj.nSamplesInBuffer = (playrec('getLastFinishedPage')-...
                obj.lastReadPage)*obj.nSamples;
            obj.lastReadPage = playrec('getLastFinishedPage');
            playrec('delPage');
            playrec('reset');
            delete(obj.t);
            status = 0;
        end % of function ImpactAcq
        
        function status = ImpactClose(obj)
            % Close impact testing module
            
            obj.nSamplesInBuffer = 0;
            obj.playChanList = [];
            obj.recChanList = [];
            status = 0;
        end % end of function ImpactClose
    end % of methods
    
    methods
        function n = get.nInCh(obj)
            n = length(obj.recChanList);
        end
        
        function n = get.nOutCh(obj)
            n = length(obj.playChanList);
        end
    end
end % of class CPlayrec
