%CTRANSFER Nihu.Lab module for Transfer Function Measurement
classdef CTransfer < CUiTab & CChSetupSensitive
    properties
        % handles to objects shared by different Modules
        Hardware        % handle to CHardware object
        AcqSetup        % handle to CAcqSetup object
        Excitation      % handle to CExcitation object
        Average         % handle to CAverage object
        LevelMeter      % handle to CLevelMeter object
        FileSelect      % handle to File Selector object
        RefSelect       % handle to Reference file select object
        
        % View objects assigned to the shared objects
        AcqSetupView    % handle to CAcqSetupView object
        ExcitationView  % handle to CExcitationView object
        AverageView     % handle to CAverageView object
        FileSelectView  % handle to File Select View object
        RefSelectView   % handle to Reference Select View object
        
        MeasurementResult   % measurement results are stored in this structure
        
        ExcData         % CGraphData object (excitation time history)
        RespData        % CGraphData object (response time history)
        IRData          % CGraphData object (Impulse response)
        FRFData         % CGraphData object (FRF)
        CohData         % CGraphData object (Coherence)
        
        ExcDataView     % handle to CGraphDataView object
        RespDataView    % handle to CGraphDataView object
        IRDataView      % handle to CGraphDataView object
        FRFDataView     % handle to CGraphDataView object
        CohDataView     % handle to CGraphDataView object
        
        hStartButton    % handle to Start Button
    end
    
    methods
        function obj = CTransfer(ParentGroup, hardware, chsetup, acqsetup,...
                excitation, average, levmeter)
            %CTRANSFER constructor of the CTransfer module
            
            obj = obj@CUiTab(ParentGroup, 'TRANSFER_MODULE');
            obj = obj@CChSetupSensitive(chsetup);
            
            % store common object handles
            obj.Hardware = hardware;
            obj.AcqSetup = acqsetup;
            obj.Excitation = excitation;
            obj.Average = average;
            obj.LevelMeter = levmeter;
            
            % instantiate CFileSelect and CReferenceSelect
            obj.FileSelect = CFileSelect({'mat'});
            obj.RefSelect = CReferenceSelect();
            
            % instantiate Graph objects
            obj.ExcData = CGraphData();
            obj.RespData = CGraphData();
            obj.IRData = CGraphData();
            obj.FRFData = CGraphData();
            obj.CohData = CGraphData();
            
            % draw the GUI module
            obj.InitFigure();
        end
        
        function SaveCallback(obj, hSaveButton)
		%SAVECALLBACK called when the user clicks on Save button

			Result = obj.MeasurementResult; %#ok<NASGU>
            fname = obj.FileSelect.getFullFileName();
            save(fname, 'Result');
            set(hSaveButton, 'BackGroundColor', get(hSaveButton, 'userData'));
        end
        
        function TransferCallback(obj, ~)
            %TRANSFERCALLBACK called when the user clicks on Start Transfer
            
            % get current parameters from Setup objects
            type = obj.Excitation.Type;         % type of excitation
            Fs = obj.AcqSetup.Fs;               % sampling frequency
            nPre = obj.Average.nPre;            % number of pre cycles
            nAverage = obj.Average.nAverage;    % number of averages
            nPost = obj.Average.nPost;          % number of post cycles
            
            % warn the user if measurement is done with mls or noise and no
            % pre cycles have been selected. The used can abort measurement
            % in this case.
            if ismember(lower(type), {'mls', 'noise'}) && nPre == 0
                button = questdlg(...
                    sprintf(CBabel.getInstance().getText('PRECYCLE_WARNING'), type),...
                    CBabel.getInstance().getText('TRANSFER_WARNING'),...
                    CBabel.getInstance().getText('YES'),...
                    CBabel.getInstance().getText('NO'),...
                    CBabel.getInstance().getText('NO'));
                drawnow(); pause(5e-2); % pause is needed because of a Matlab GUI bug
                if strcmpi(button, CBabel.getInstance().getText('NO'))
                    return;
                end
            end
            
            obj.blockUI();
            
            % reset cycle counter
            obj.Average.reset();
            
            % generate excitation
            x = obj.generateExcitationSignal();
            
            try
                % perfom measurement, obtain results as Voltage
                % yRaw does not contain the response to post and pre cycles
                [~, yRaw] = obj.Hardware.TransferAcq(nPre, nAverage, nPost, Fs, x);
            catch ME
                errordlg(sprintf('%s\n%s\n%s:%d',...
                    ME.identifier, ME.message, ME.stack(1).name,...
                    ME.stack(1).line), '!!! Error !!!', 'modal');
            end
            
            % discard pre and post excitation cycles
            if strcmpi(type, 'noise')
                x = x(:,nPre+(1:nAverage));
            end
            
            % display levels on LevelMeter
            obj.LevelMeter.UpdateLevels(squeeze(max(yRaw,[],2)));
            
            % take channel sensitivity into account, compute EU
            y = obj.ChannelSetup.voltage2eu(yRaw);
            y = obj.ChannelSetup.decompensate9234(y, Fs);
            
			% compute transfer functions and impulse response
            [tScale, x, h, fScale, H, C] = obj.computeTransfer(x, y);
            
            % display result graphs
            obj.displayGraphs(tScale, x(:,end), y(:,end,:), h,...
                fScale, abs(H), abs(C));
            
            % store measurement results into object memory
            obj.storeMeasurementData(x, y, h, H, C);
            
            % find next filename
            obj.FileSelect.NextAvailable();
            % mark save button with green
            handles = guihandles(gcbo);
            hSave = handles.SAVE_TRANSFER;
            set(hSave, 'BackGroundColor', 'green');
            
            obj.unblockUI()
        end % of function TransferCallback

    end % of public methods
    
    methods (Access = private)
        function InitFigure(obj)
            %INITFIGURE draw the Module to the GUI.
            % This function is called only once by the constructor
            
            % create View classes
            obj.AcqSetupView = CAcqSetupView(obj.hContainer, obj.AcqSetup);
            obj.ExcitationView = CExcitationView(obj.hContainer, obj.Excitation);
            obj.AverageView = CAverageView(obj.hContainer, obj.Average);
            h2 = CLevelMeterView(obj.hContainer, obj.LevelMeter, 100);
            obj.FileSelectView = CFileSelectView(obj.hContainer, obj.FileSelect);
            obj.RefSelectView = CReferenceSelectView(obj.hContainer, obj.RefSelect);
            % create start/stop acquisition button
            obj.hStartButton = uicontrol('parent', obj.hContainer,...
                'Style', 'PushButton',...
                'Tag', 'START_TRANSFER', ...
                'String', CBabel.getInstance().getText('START_TRANSFER'), ...
                'Callback', @(src, event)TransferCallback(obj, src));
            % create save button
            butSav = uicontrol('parent', obj.hContainer,...
                'Style', 'PushButton',...
                'Tag', 'SAVE_TRANSFER',...
                'String', CBabel.getInstance().getText('SAVE_TRANSFER'), ...
                'Callback', @(src, event)SaveCallback(obj, src));
            % store default background color into butSav
            % this is needed to be able to revert to the default color
            % after the Transfer Data have been saved
            set(butSav, 'userData', get(butSav, 'BackGroundColor'));
            
            h3 = CHardwareView(obj.hContainer, obj.Hardware);
            obj.IRDataView = CGraphDataView(obj.hContainer, obj.IRData, 'IMPULSE_RESPONSE');
            obj.FRFDataView = CGraphDataView(obj.hContainer, obj.FRFData, 'FREQUENCY_RESPONSE_FUNCTION');
            obj.ExcDataView = CGraphDataView(obj.hContainer, obj.ExcData, 'EXCITATION_TIME_HISTORY');
            obj.RespDataView = CGraphDataView(obj.hContainer, obj.RespData, 'RESPONSE_TIME_HISTORY');
            obj.CohDataView = CGraphDataView(obj.hContainer, obj.CohData, 'COHERENCE_FUNCTION');
            
            % get absolute width of the AcqSetupView panel
            w = get(obj.AcqSetupView.hPanel, 'position');
            w = w(3);
            
            % arrange GUI elements
            lMenu = {{obj.AcqSetupView.hPanel, [-1 0]};
                {obj.ExcitationView.hPanel, [-1 0]};
                {obj.AverageView.hPanel, [-1 0]};
                {obj.FileSelectView.hPanel, [-1 0]};
                {obj.RefSelectView.hPanel, [-1 0]};
                {butSav, [-1 CSetupClass.itemHeight]};
                {obj.hStartButton, [-1 CSetupClass.itemHeight]};
                {h2.hPanel, [-1 -1]};
                {h3.hPanel, [-1 0]}};
            content = {{{{obj.ExcDataView.hPanel, [-1 -1]};
                {obj.RespDataView.hPanel, [-1 -1]};
                {obj.IRDataView.hPanel, [-1 -1]}}, [-1 -1]} {{{obj.FRFDataView.hPanel, [-1 -1]};
                {obj.CohDataView.hPanel, [-1 -1]}}, [-1 -1]}};
            Layout = {{lMenu, [w, -1]} {content, [-1 -1]}};
            GuiLayout.ApplyLayout(obj.hContainer, Layout);
            
            movegui(obj.hContainer, 'center');
            movegui(obj.hContainer, 'onscreen');
        end
        
        function x = generateExcitationSignal(obj)
            %GENERATEEXCITATIONSIGNAL
            
            Fs = obj.AcqSetup.Fs;               % sampling frequency
            nSamples = obj.AcqSetup.nSamples;   % number of samples
            nPre = obj.Average.nPre;            % number of pre cycles
            nAverage = obj.Average.nAverage;    % number of averages
            nPost = obj.Average.nPost;          % number of post cycles
            nTot = nPre + nAverage + nPost;     % total number of cycles
            
            switch lower(obj.Excitation.Type)
                case {'harmonic', 'mls', 'sweep'}
                    % Only one cycle is needed
                    x = obj.Excitation.GenerateSignal(Fs, nSamples);
                case {'noise'}
                    % different noise realisations are generated in each
                    % column of x for each measurement cycle
                    x = reshape(obj.Excitation.GenerateSignal(Fs, nTot*nSamples), nSamples, nTot);
            end
        end
        
        function [tScale, x, h, fScale, H, C] = computeTransfer(obj, x, y)
            %COMPUTETRANSFER
            
            nSamples = size(x,1);
            Fs = obj.AcqSetup.Fs;
            tScale = (0 : nSamples-1)' / Fs;        % time scale
            fScale = (0:nSamples-1)'/nSamples*Fs;   % frequency scale
            freqIndex = fScale <= Fs/2;                 % limit fscale to fs/2
            fScale = fScale(freqIndex);
            
            % if reference channel was used, x needs to be replaced by yref
            refind = obj.ChannelSetup.getActiveReferenceIndex();
            if ~isempty(refind)
                x = y(:,:,refind);
            end
            
            switch lower(obj.Excitation.Type)
                case 'harmonic'
                    [H, C] = transferest(x, y, 'harmonic', obj.Excitation.Frequency/Fs);
                    obj.DispHarmonicTransfer(H, C);
                    H = nan(length(fScale), size(H,2));
                    C = nan(length(fScale), size(H,2));
                    h = nan(nSamples, size(H,2));
                otherwise
                    if ~isempty(refind)
                        [H, C, h] = transferest(x, y, 'noise');
                    else
                        [H, C, h] = transferest(x, y, obj.Excitation.Type);
                    end
                    
                    H = H(freqIndex,:);
                    C = C(freqIndex,:);
                    
                    % Check if external reference was used
                    if obj.RefSelect.IsActive
                        [H0, C0] = obj.RefSelect.GetReferenceFRF(Fs, nSamples);
                        H = bsxfun(@times, H, 1./H0);
                        C = bsxfun(@times, C, C0);
                    end
            end
        end
        
        function displayGraphs(obj, tScale, x, y, h, fScale, H, C)
            %DISPLAYGRAPHS
            
            obj.ExcData.UpdateData(tScale, squeeze(x));
            obj.RespData.UpdateData(tScale, squeeze(y));
            obj.IRData.UpdateData(tScale, h);
            obj.FRFData.UpdateData(fScale, H);
            obj.CohData.UpdateData(fScale, C);
        end
        
        function storeMeasurementData(obj, x, y, h, H, C)
            %STOREMEASUREMENTDATA
            
            obj.MeasurementResult.Channels = obj.ChannelSetup.getUsedChannels();
            obj.MeasurementResult.AcqSetup = obj.AcqSetup.dcopy();
            obj.MeasurementResult.Excitation = obj.Excitation.dcopy();
            obj.MeasurementResult.Average = obj.Average.dcopy();
            obj.MeasurementResult.ExcitationSignal = x;
            obj.MeasurementResult.ResponseSignal = y;
            obj.MeasurementResult.ImpulseResponse = h;
            obj.MeasurementResult.FrequencyResponseFunction = H;
            obj.MeasurementResult.Coherence = C;
        end
        
        function blockUI(obj)
            %BLOCKUI
            
            set(gcbf, 'UserData', get(gcbf, 'CloseRequestFcn'));
            set(gcbf, 'CloseRequestFcn', '');
            obj.hParentGroup.block();
            obj.AcqSetupView.block();
            obj.ExcitationView.block();
            obj.AverageView.block();
            set(obj.hStartButton, 'Enable', 'off');
        end
        
        function unblockUI(obj)
            %UNBLOCKUI
            
            set(gcbf, 'CloseRequestFcn', get(gcbf, 'UserData'));
            obj.hParentGroup.unblock();
            obj.AcqSetupView.unblock();
            obj.ExcitationView.unblock();
            obj.AverageView.unblock();
            set(obj.hStartButton, 'Enable', 'on');
        end
        
        function DispHarmonicTransfer(obj, Amp, Coh)
            MainFigure = figure('Name', 'Harmonic Measurement Results', ...
                'Units', 'pixels', ...
                'MenuBar', 'None',...
                'NumberTitle', 'off',...
                'Resize', 'on', ...
                'Color', get(0,'defaultUicontrolBackgroundColor'), ...
                'WindowStyle', 'modal',...
                'Visible', 'off');
            
            Ch = obj.ChannelSetup.InputChannels(cell2mat({obj.ChannelSetup.InputChannels.isSelected}));
            nCh = length(Ch);
            data = cell(nCh,5);
            for iCh = 1 : nCh
                data(iCh,:) = {Ch(iCh).Name, abs(Amp(iCh)),...
                    sprintf('%s/V', Ch(iCh).Sensor.Quantity.EU), angle(Amp(iCh))/pi*180, Coh(iCh)};
            end
            table = uitable('parent', gcf,...
                'Units', 'pixels',...
                'Data', data,...
                'ColumnName', {'Name', 'Amplitude', 'EU', 'Phase [o]', 'Coherence [-]'},...
                'ColumnWidth', {100, 80, 60, 60, 60},...
                'RowName', {Ch.PhysChNr},...
                'ColumnFormat', {'char', 'numeric', 'char', 'numeric', 'numeric'},...
                'ColumnEditable', false(1,6));
            set(MainFigure, 'Visible', 'on');
            ex = get(table, 'extent');
            set(table, 'position', ex);
            GuiLayout.ApplyLayout(MainFigure, {{table, [0 0]}});
        end
    end % of private methods
    
    methods
        function show(obj)
            %SHOW Called when the user opens this GUI module
            
            obj.LevelMeter.Reset();
            
            if obj.chSetupChanged
                Chan = obj.ChannelSetup.getUsedChannels();
                Col = {Chan.Color};
                Col = cell2mat(Col(:));
                
                obj.IRData.UpdateChannels(cell2mat({Chan.PhysChNr}),...
                    {Chan.Name},...
                    CQuantity.Time.textID(), CQuantity.Time.EU,...
                    Chan.getQuantText(), Chan.getEUText(), Col);
                obj.IRDataView.UpdateChannels();
                
                obj.FRFData.UpdateChannels(cell2mat({Chan.PhysChNr}),...
                    {Chan.Name},...
                    CQuantity.Frequency.textID(), CQuantity.Frequency.EU,...
                    Chan.getQuantText(), Chan.getEUText(), Col);
                obj.FRFDataView.UpdateChannels();
                
                obj.ExcData.UpdateChannels(0,...
                    {CBabel.getInstance().getText('EXCITATION')},...
                    CQuantity.Time.textID(), CQuantity.Time.EU,...
                    {CQuantity.Voltage.textID()}, {CQuantity.Voltage.EU}, Col(1,:));
                obj.ExcDataView.UpdateChannels();
                
                obj.RespData.UpdateChannels(cell2mat({Chan.PhysChNr}),...
                    {Chan.Name},...
                    CQuantity.Time.textID(), CQuantity.Time.EU,...
                    Chan.getQuantText(), Chan.getEUText(), Col);
                obj.RespDataView.UpdateChannels();
                
                obj.CohData.UpdateChannels(cell2mat({Chan.PhysChNr}),...
                    {Chan.Name},...
                    CQuantity.Frequency.textID(), CQuantity.Frequency.EU,...
                    CQuantity.Dimensionless.textID(), CQuantity.Dimensionless.EU, Col);
                
                obj.chSetupChanged = false;
            end
            
            obj.CohDataView.UpdateChannels();
            obj.IRDataView.unblock();
            obj.FRFDataView.unblock();
            obj.ExcDataView.unblock();
            obj.RespDataView.unblock();
            obj.CohDataView.unblock();
            
            status = obj.Hardware.TransferInit(obj.ChannelSetup.InputChannels, obj.ChannelSetup.OutputChannels);
        end
        
        function hide(obj)
            %HIDE function called each time before the module is closed
            
            obj.IRDataView.block();
            obj.FRFDataView.block();
            obj.CohDataView.block();
            obj.ExcDataView.block();
            obj.RespDataView.block();
            
            status = obj.Hardware.TransferClose();
        end
    end % of public methods
end
