function [A, Freq, init, T0] = LevelPhase(x, fs, T0, f0, type, init)
%LEVELPHASE  Level and phase of a harmonic signal component
%   A = LevelPhase(X, FS, T0, F0, TYPE) determines the amplitude and phase of
%   the signal X at a given frequency F. The algorithm is a vector voltage
%   probe, where the signal X is projected to a cosine and a sine signal
%   to obtain its vector components. The parameter TYPE determines the low
%   pass filter algorithm. If TYPE is 'exact' (default), the LPF is
%   performed by means of averaging. If TYPE is 'butter', then the low
%   pass filter is a Butterworth filter of order 5 and cutoff frequency
%   10 Hz.

% Copyright 2010-2011 Peter Fiala

% default type argument
if nargin < 5
    type = 'exact';
elseif nargin < 6
    init = zeros(5,1);
end

s = size(x);
nSamples = s(1);   % number of samples
x = reshape(x, nSamples, []);
nCh = size(x,2); % remaining size
t = T0+(0:nSamples-1)'/fs;
X = repmat(2*exp((-1i*2*pi*f0)*t), [1, nCh]) .* x;

switch type
    case 'exact'
        A = mean(X,1);
    case 'butter'
        fc = 10;
        [b, a] = butter(5, max(fc/(fs/2), 2e-3));
        [X, init] = filter(b, a, X, init, 1);
        [A, gamma] = expfit(X, fs);
        Freq = f0 + imag(gamma)/2/pi;
    otherwise
        error('Unknown filter type %s', type);
end
n = s(2:end);
if isscalar(n)
    n = [n 1];
end
A = reshape(A, n);
T0 = T0 + nSamples/fs;
end
