function mls = mlssignal(N)
%MLSSIGNAL Maximum Length Sequence
%   MLS = MLSSIGNAL(N)   Generates Maximum Length Sequence MLS of order N.
%   The MLS signal is returned in the the column vectror MLS. The MLS
%   signal is a binary signal with values -1 and +1.
%
% Example
%   x = mlssignal(16);
%   soundsc(x, 44100);
%
% See also SWEEPSIGNAL

% Copyright 2010-2010 Peter Fiala

% parameter check
if N > 32 || N < 1 || floor(N) ~= N
    error('NiHuLab:Param',...
        'The MLS order N must be an integer value between 1 and 32');
end

try % mex version (fast)
    mls = double(mlssignal_mex(N));
catch % matlab version (very slow)
    % MLS masks from N = 1 to N = 32
    Taps = uint32([1, 3, 6, 12, 20, 48, 72, 184, 272, 576, 1280, 3232, 6912,...
        12424, 24576, 53256, 73728, 132096, 405504, 589824, 1048578,...
        2097153, 4194320, 8388621, 16777220, 33554625, 67109057, 134217732,...
        268435458, 536920065, 1073741828, 2348810241]);
    mask = Taps(N); % select MLS mask
    
    M = 2^N-1;  % length of MLS signal
    mls = zeros(M,1);   % allocate space for output signal
    
    m = uint32(Inf); % initial state of shift reg. (11111111...111)
    for ind = 1:M
        b = mod(sum(bitget(bitand(m,mask),1:N)),2);
        m = bitshift(m,1)+b;
        mls(ind) = b;
    end
end
mls = mls - ~mls; % symmetrize MLS signal
end
