/**
 * \file nihu_daqmx_mex.c
 * \brief MEX interface for DAQmx devices
 */

#include <mex.h>
#include <string.h>
#include "nihu_daqmx.h"
#include <stdint.h>

/** \brief call DAQmx function and throw Matlab error or warning if needed */
#define DAQMX_CALL(func) if ((DAQerr = func) != 0) mex_daqmx_error(DAQerr)

/** \brief type of a preprocessed MEX function */
typedef void (*mxFunction)(mxArray **, mxArray const **);

/** \brief generate a Matlab error or warning based on a DAQmx error value
 * \param [in] err the DAQmx error value
 */
void mex_daqmx_error(int32 err)
{
	char extErrorBuf[MAX_STR];
	char errorBuf[MAX_STR];
	DAQmxGetExtendedErrorInfo(extErrorBuf, MAX_STR);
	DAQmxGetErrorString(err, errorBuf, MAX_STR);
	if (err < 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:DAQerror", "DAQmx: %d: %s\n%s", err,
			errorBuf, extErrorBuf);
	else if (err > 0)
		mexWarnMsgIdAndTxt("NiHuDAQmx:DAQwarning", "DAQmx: %d: %s\n%s", err,
			errorBuf, extErrorBuf);
}

/** \brief convert device properties to a MEX structure array
 * \param [in] devProp array of device properties
 * \param [in] n length of the array
 * \return a MEX array containing the structure array
 */
mxArray *read_device_properties(DeviceProperties const devProp[], int n)
{
	int i;

	char const *fieldNames[] = {
		"DeviceName",
		"ProductNumber",
		"ProductType",
		"ProductCategory",
		"SerialNumber"
	};

	/* create array of structures */
	mxArray *array = mxCreateStructMatrix(n, 1,
		sizeof(fieldNames)/sizeof(fieldNames[0]), fieldNames);

	for (i = 0; i < n; ++i)
	{
		mxArray *elem;

		/* copy device name */
		mxSetField(array, i, "DeviceName",
			mxCreateString(devProp[i].deviceName));

		/* copy product number */
		elem = mxCreateNumericMatrix(1, 1, mxUINT64_CLASS, mxREAL);
		*((uint64_t*)mxGetData(elem)) = devProp[i].productNumber;
		mxSetField(array, i, "ProductNumber", elem);

		/* copy product type */
		mxSetField(array, i, "ProductType",
			mxCreateString(devProp[i].productType));

		/* copy product category */
		elem = mxCreateNumericMatrix(1, 1, mxUINT64_CLASS, mxREAL);
		*((uint64_t*)mxGetData(elem)) = devProp[i].productCategory;
		mxSetField(array, i, "ProductCategory", elem);

		/* copy product serial number */
		elem = mxCreateNumericMatrix(1, 1, mxUINT64_CLASS, mxREAL);
		*((uint64_t*)mxGetData(elem)) = devProp[i].serialNumber;
		mxSetField(array, i, "SerialNumber", elem);
	}

	return array;
}

/** \brief convert chassis properties to a MEX structure
 * \param [in] chasProp pointer to chassis properties structure
 * \return a MEX array containing the chassis properties structure
 */
mxArray *read_chassis_properties(ChassisProperties const *chasProp)
{
	char const *fieldNames[] = {
		"ChassisProperties",
		"Modules"
	};

	/* create single array */
	mxArray *array = mxCreateStructMatrix(1, 1,
		sizeof(fieldNames)/sizeof(fieldNames[0]), fieldNames);

	/* copy device properties */
	mxSetField(array, 0, "ChassisProperties",
		read_device_properties(&chasProp->chassisProperties, 1));
	/* copy module property array */
	mxSetField(array, 0, "Modules",
		read_device_properties(chasProp->moduleDeviceProperties,
		chasProp->numModules));

	return array;
}


void get_device_names(mxArray *plhs[], mxArray const *prhs[])
{
	int32 DAQerr;
	char dev_names[MAX_STR];
	DAQMX_CALL(DAQmxGetSysDevNames(dev_names, MAX_STR));
	plhs[0] = mxCreateString(dev_names);
}


/** \brief create a new DAQmx task
 * \details Call this function from Matlab as
 *
 * task_handle = NiHuDAQmx('create_task', task_name);
 */
void create_task(mxArray *plhs[], mxArray const *prhs[])
{
	TaskHandle taskHandle;
	int32 DAQerr;
	char task_name[MAX_STR];
	int err;

	if (err = mxGetString(prhs[0], task_name, MAX_STR) != 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:param", "Invalid task name");
	DAQMX_CALL(DAQmxCreateTask(task_name, &taskHandle));

	plhs[0] = mxCreateNumericMatrix(1, 1, mxUINT64_CLASS, mxREAL);
	*((uint64_t*)mxGetData(plhs[0])) = (uint64_t)taskHandle;
}

/** \brief start a DAQmx task
 * \details Call this function from Matlab as
 *
 * NiHuDAQmx('start_task', task_handle);
 *
 * if task_handles is a vector then the tasks are started synchronized
 */
void start_task(mxArray *plhs[], mxArray const *prhs[])
{
	uint64_t *taskHandles;
	int32 DAQerr;
	int nTasks, i;

	taskHandles = (uint64_t*)mxGetData(prhs[0]);
	nTasks = mxGetM(prhs[0]) * mxGetN(prhs[0]);
	if (nTasks > 1)
	{
		char triggerSource[MAX_STR] = "/";					// the trigger source must start with a slash
		TaskHandle hin, hout;

		hin	= (TaskHandle)taskHandles[0];			// input is 0-th
		hout = (TaskHandle)taskHandles[1];			// output is 1-st
		
		// get input start trigger name
		DAQMX_CALL(DAQmxGetNthTaskDevice(hin, 1, triggerSource+1, MAX_STR-1));
		strcat(triggerSource, "/ai/StartTrigger");
		
		DAQMX_CALL(DAQmxCfgDigEdgeStartTrig(hout, triggerSource, DAQmx_Val_Rising));

		// start output and input in this order
		DAQMX_CALL(DAQmxStartTask(hout));
		DAQMX_CALL(DAQmxStartTask(hin));
	}
	else
	{
		DAQMX_CALL(DAQmxStartTask((TaskHandle)taskHandles[0]));
	}
}

/** \brief stop a DAQmx task
 * \details Call this function from Matlab as
 *
 * NiHuDAQmx('stop_task', task_handles);
 */
void stop_task(mxArray *plhs[], mxArray const *prhs[])
{
	uint64_t *taskHandles;
	int32 DAQerr;
	int nTasks, i;

	nTasks = mxGetM(prhs[0]) * mxGetN(prhs[0]);
	taskHandles = (uint64_t*)mxGetData(prhs[0]);
	for (i = 0; i < nTasks; ++i)
		DAQMX_CALL(DAQmxStopTask((TaskHandle)taskHandles[i]));
}

/** \brief clear a DAQmx task
 * \details Call this function from Matlab as
 *
 * NiHuDAQmx('clear_task', task_handles);
 */
void clear_task(mxArray *plhs[], mxArray const *prhs[])
{
	uint64_t *taskHandles;
	int32 DAQerr;
	int nTasks, i;

	nTasks = mxGetM(prhs[0]) * mxGetN(prhs[0]);
	taskHandles = (uint64_t*)mxGetData(prhs[0]);
	for (i = 0; i < nTasks; ++i)
		DAQMX_CALL(DAQmxClearTask((TaskHandle)taskHandles[i]));
}

/** \brief return a task's name
 * \details Call this function from Matlab as
 *
 * task_name = NiHuDAQmx('get_task_name', task_handle);
 */
void get_task_name(mxArray *plhs[], mxArray const *prhs[])
{
	TaskHandle task_handle;
	int32 DAQerr;
	char task_name[MAX_STR];

	task_handle = (TaskHandle) *((uint64_t*)mxGetData(prhs[0]));
	DAQMX_CALL(DAQmxGetTaskName(task_handle, task_name, MAX_STR));
	plhs[0] = mxCreateString(task_name);
}

/** \brief return a task's channels
 * \details Call this function from Matlab as
 *
 * task_channels = NiHuDAQmx('get_task_channels', task_handle);
 */
void get_task_channels(mxArray *plhs[], mxArray const *prhs[])
{
	TaskHandle task_handle;
	int32 DAQerr;
	char task_channels[MAX_STR];

	task_handle = (TaskHandle) *((uint64_t*)mxGetData(prhs[0]));
	DAQMX_CALL(DAQmxGetTaskChannels(task_handle, task_channels, MAX_STR));
	plhs[0] = mxCreateString(task_channels);
}

/** \brief return the number of a task's channels
 * \details Call this function from Matlab as
 *
 * num_channels = NiHuDAQmx('get_task_num_channels', task_handle);
 */
void get_task_num_channels(mxArray *plhs[], mxArray const *prhs[])
{
	TaskHandle task_handle;
	int32 DAQerr;
	int32 num_channels;

	task_handle = (TaskHandle) *((uint64_t*)mxGetData(prhs[0]));
	DAQMX_CALL(DAQmxGetTaskNumChans(task_handle, &num_channels));
	plhs[0] = mxCreateNumericMatrix(1, 1, mxINT32_CLASS, mxREAL);
	*(int32*)mxGetData(plhs[0]) = num_channels;
}

/** \brief reserve a network device by name
 * \details Call this function from Matlab as
 *
 * NiHuDAQmx('reserve_network_device', device_name);
 */
void reserve_network_device(mxArray *plhs[], mxArray const *prhs[])
{
	char device_name[MAX_STR];
	int32 DAQerr;
	int err;

	err = mxGetString(prhs[0], device_name, MAX_STR);
	if (err != 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:param", "Invalid network device name");

	DAQMX_CALL(DAQmxReserveNetworkDevice(device_name, 1));
}

/** \brief unreserve a network device by name
 * \details Call this function from Matlab as
 *
 * NiHuDAQmx('unreserve_network_device', device_name);
 */
void unreserve_network_device(mxArray *plhs[], mxArray const *prhs[])
{
	char device_name[MAX_STR];
	int32 DAQerr;
	int err;

	err = mxGetString(prhs[0], device_name, MAX_STR);
	if (err != 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:param", "Invalid network device name");

	DAQMX_CALL(DAQmxUnreserveNetworkDevice(device_name));
}

/** \brief get structure containing device properties
 * \details Call this function from Matlab as
 *
 * prop = NiHuDAQmx('get_device_properties', device_name);
 */
void get_device_properties(mxArray *plhs[], mxArray const *prhs[])
{
	char device_name[MAX_STR];
	int err;
	int32 DAQerr;
	DeviceProperties device_properties;

	err = mxGetString(prhs[0], device_name, MAX_STR);
	if (err != 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:param", "Invalid device name");

	DAQMX_CALL(getDeviceProperties(device_name, &device_properties));

	plhs[0] = read_device_properties(&device_properties, 1);
}



/** \brief get structure containing chassis properties
 * \details Call this function from Matlab as
 *
 * prop = NiHuDAQmx('get_chassis_properties', device_name);
 */
void get_chassis_properties(mxArray *plhs[], mxArray const *prhs[])
{
	char device_name[MAX_STR];
	int err;
	int32 DAQerr;
	ChassisProperties chassis_properties;

	err = mxGetString(prhs[0], device_name, MAX_STR);
	if (err != 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:param", "Invalid device name");

	DAQMX_CALL(getChassisProperties(device_name, &chassis_properties));

	plhs[0] = read_chassis_properties(&chassis_properties);
}

/** \brief create an analogue input channel
 * \details Call this function from Matlab as
 *
 * NiHuDAQmx('create_input_channel', task_handle, channel_name, range, iepe);
 */
void create_input_channel(mxArray *plhs[], mxArray const *prhs[])
{
	int err;
	int32 DAQerr;
	TaskHandle task_handle;
	char channel_name[MAX_STR];
	float64 min_val, max_val;
	int iepe;
	
	task_handle = (TaskHandle) *((uint64_t*)mxGetData(prhs[0]));
	err = mxGetString(prhs[1], channel_name, MAX_STR);
	if (err != 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:param", "Invalid channel name");
		
	min_val = (float64)(*mxGetPr(prhs[2]));
	max_val = (float64)(*(mxGetPr(prhs[2]) + 1));
	iepe = (int) (*(mxLogical *)mxGetData(prhs[3]));
	
	DAQMX_CALL(DAQmxCreateAIVoltageChan(task_handle, channel_name, NULL,
		DAQmx_Val_Cfg_Default, min_val, max_val, DAQmx_Val_Volts, NULL));
		
	if (iepe)
	{
		DAQMX_CALL(DAQmxSetAICoupling(task_handle, channel_name, DAQmx_Val_AC));
		DAQMX_CALL(DAQmxSetAIExcitVoltageOrCurrent(task_handle, channel_name,
			DAQmx_Val_Current));
		DAQMX_CALL(DAQmxSetAIExcitSrc(task_handle, channel_name,
			DAQmx_Val_Internal));
		DAQMX_CALL(DAQmxSetAIExcitVal(task_handle, channel_name, 2e-3));
	}
}

/** \brief create an analogue output channel
 * \details Call this function from Matlab as
 *
 * NiHuDAQmx('create_output_channel', task_handle, channel_name, range);
 */
void create_output_channel(mxArray *plhs[], mxArray const *prhs[])
{
	int err;
	int32 DAQerr;
	TaskHandle task_handle;
	char channel_name[MAX_STR];
	float64 min_val, max_val;
	
	task_handle = (TaskHandle) *((uint64_t*)mxGetData(prhs[0]));
	err = mxGetString(prhs[1], channel_name, MAX_STR);
	if (err != 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:param", "Invalid channel name");
		
	min_val = (float64)(*mxGetPr(prhs[2]));
	max_val = (float64)(*(mxGetPr(prhs[2]) + 1));
	
	DAQMX_CALL(DAQmxCreateAOVoltageChan(task_handle, channel_name, NULL,
		min_val, max_val, DAQmx_Val_Volts, NULL));
}

/** \brief configure timing and sampling parameters
 * \details Call this function from Matlab as
 *
 * NiHuDAQmx('configure_timing', task_handles, sample_rate, is_continuous, num_samples_per_ch);
 *
 * For continuous sampling, the last parameter is the buffer size per channel.
 * For synchronous timing, the first task handle should be the input task.
 */
void configure_timing(mxArray *plhs[], mxArray const *prhs[])
{
	int32 DAQerr;
	uint64_t *taskHandles;
	int numTasks, i;
	float64 sample_rate;
	uInt64 num_samples;
	int continuous;

	taskHandles = (uint64_t*)mxGetData(prhs[0]);
	numTasks = mxGetN(prhs[0]) * mxGetM(prhs[0]);
	sample_rate = (float64)(*mxGetPr(prhs[1]));
	continuous =  (int) *((mxLogical *)mxGetData(prhs[2]));
	num_samples = *((uint64_t*)mxGetData(prhs[3]));

	for (i = 0; i < numTasks; ++i)
		DAQMX_CALL(DAQmxCfgSampClkTiming(taskHandles[i],
			i == 0 ? NULL : "ai/SampleClock",
			sample_rate,
			DAQmx_Val_Rising,
			continuous ? DAQmx_Val_ContSamps : DAQmx_Val_FiniteSamps,
			num_samples));
}

/** \brief read input data from buffer
 * \details Call this function from Matlab as
 *
 * [data, num_read] = NiHuDAQmx('read', task_handle, num_samples_per_channel);
 */
void read(mxArray *plhs[], mxArray const *prhs[])
{
	int32 DAQerr;
	int32 num_samples;
	int32 num_channels;
	TaskHandle task_handle;
	
	task_handle = (TaskHandle) *((uint64_t*)mxGetData(prhs[0]));
	num_samples = *((int32_t*)mxGetData(prhs[1]));

	DAQMX_CALL(DAQmxGetTaskNumChans(task_handle, &num_channels));

	plhs[0] = mxCreateNumericMatrix(num_samples, num_channels,
		mxDOUBLE_CLASS, mxREAL);
	plhs[1] = mxCreateNumericMatrix(1, 1, mxINT32_CLASS, mxREAL);
			
	DAQMX_CALL(DAQmxReadAnalogF64(task_handle, num_samples,
		DAQmx_Val_WaitInfinitely,
		DAQmx_Val_GroupByChannel, mxGetPr(plhs[0]), num_samples*num_channels,
		(int32*)mxGetData(plhs[1]), NULL));
}


/** \brief write data to output buffer
 * \details Call this function from Matlab as
 *
 * num_written = NiHuDAQmx('write', task_handle, data);
 */
void write(mxArray *plhs[], mxArray const *prhs[])
{
	int32 DAQerr;
	int32 num_samples_per_chan;
	TaskHandle task_handle;
	
	task_handle = (TaskHandle) *((uint64_t*)mxGetData(prhs[0]));
	num_samples_per_chan = mxGetM(prhs[1]);

	plhs[0] = mxCreateNumericMatrix(1, 1, mxINT32_CLASS, mxREAL);
			
	DAQMX_CALL(DAQmxWriteAnalogF64(task_handle, num_samples_per_chan, (bool32)0,
		DAQmx_Val_WaitInfinitely, DAQmx_Val_GroupByChannel,
		mxGetPr(prhs[1]), (int32*)mxGetData(plhs[0]), NULL));
}


/** \brief the entry point of the MEX interface
 * \param [in] nlhs number of left hand side arguments
 * \param [out] plhs pointer array to left hand side arguments
 * \param [in] nrhs number of right hand side arguments
 * \param [in] prhs pointer array to right hand side arguments
 */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, mxArray const *prhs[])
{
	int i;
	char option[MAX_STR];
	int err;

	struct {
		char const *option;
		mxFunction func;
		int nlhs, nrhs;
	} functions[] = {
		/* dev_names = NiHuDAQmx('get_device_names') */
		{"get_device_names", get_device_names, 1, 0},
		/* task_handle = NiHuDAQmx('create_task', task_name) */
		{"create_task", create_task, 1, 1},
		/* NiHuDAQmx('start_task', task_handle) */
		{"start_task", start_task, 0, 1},
		/* NiHuDAQmx('stop_task', task_handle) */
		{"stop_task", stop_task, 0, 1},
		/* NiHuDAQmx('clear_task', task_handle) */
		{"clear_task", clear_task, 0, 1},
		/* task_name = NiHuDAQmx('get_task_name', task_handle); */
		{"get_task_name", get_task_name, 1, 1},
		/* task_channels = NiHuDAQmx('get_task_channels', task_handle); */
		{"get_task_channels", get_task_channels, 1, 1},
		/* num_channels = NiHuDAQmx('get_task_num_channels', task_handle); */
		{"get_task_num_channels", get_task_num_channels, 1, 1},
		/* NiHuDAQmx('reserve_network_device', device_name) */
		{"reserve_network_device", reserve_network_device, 0, 1},
		/* NiHuDAQmx('unreserve_network_device', device_name) */
		{"unreserve_network_device", unreserve_network_device, 0, 1},
		/* device_properties = NiHuDAQmx('get_device_properties', device_name) */
		{"get_device_properties", get_device_properties, 1, 1},
		/* chassis_properties = NiHuDAQmx('get_chassis_properties', device_name) */
		{"get_chassis_properties", get_chassis_properties, 1, 1},
		/* NiHuDAQmx('create_input_channel', task_handle, channel_name, range, iepe) */
		{"create_input_channel", create_input_channel, 0, 4},
		/* NiHuDAQmx('create_output_channel', task_handle, channel_name, range) */
		{"create_output_channel", create_output_channel, 0, 3},
		/* NiHuDAQmx('configure_timing', task_handle, sample_rate, is_continuous, num_samp_per_ch) */
		{"configure_timing", configure_timing, 0, 4},
		/* [data, num_read] = NiHuDAQmx('read', task_handle, num_samples_per_channel) */
		{"read", read, 2, 2},
		/* num_written = NiHuDAQmx('write', task_handle, data) */
		{"write", write, 1, 2},
		/* terminator */
		{"", NULL, 0, 0}
	};
	
	if (nrhs < 1 || (err = mxGetString(prhs[0], option, MAX_STR)) != 0)
		mexErrMsgIdAndTxt("NiHuDAQmx:param", "Unable to read option string");
	for (i = 0; functions[i].func != NULL; ++i)
	{
		if (strcmp(option, functions[i].option) == 0)
		{
			if (nlhs != functions[i].nlhs)
				mexErrMsgIdAndTxt("NiHuDAQmx:param",
					"Invalid number of lhs parameters: %s, %d, should be %d",
					option, nlhs, functions[i].nlhs);
			if (nrhs != functions[i].nrhs+1)
				mexErrMsgIdAndTxt("NiHuDAQmx:param",
					"Invalid number of rhs parameters: %s, %d, should be %d",
					option, nrhs, functions[i].nrhs);
			functions[i].func(plhs, prhs+1);
			return;
		}
	}
	
	mexErrMsgIdAndTxt("NiHuDAQmx:param", "Unknown option parameter: %s", option);
}
